#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os, plistlib, struct

from .loader import QlLoader

from qiling.exception import *
from qiling.const import *

from .macho_parser.parser import *
from .macho_parser.const import *
from .macho_parser.utils import *
from qiling.os.macos.kernel_api.hook import *
from qiling.os.memory import QlMemoryHeap

from qiling.os.macos.const import *
from qiling.os.macos.task import MachoTask
from qiling.os.macos.kernel_func import FileSystem, map_commpage
from qiling.os.macos.mach_port import MachPort, MachPortManager
from qiling.os.macos.subsystems import MachHostServer, MachTaskServer
from qiling.os.macos.utils import env_dict_to_array, page_align_end
from qiling.os.macos.thread import QlMachoThreadManagement, QlMachoThread


# commpage is a shared mem space which is in a static address
def load_commpage(ql):
    if ql.archtype == QL_ARCH.X8664:
        COMM_PAGE_START_ADDRESS = X8664_COMM_PAGE_START_ADDRESS
    else:    
        COMM_PAGE_START_ADDRESS = ARM64_COMM_PAGE_START_ADDRESS

    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_SIGNATURE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_CPU_CAPABILITIES64, b'\x00\x00\x00\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_UNUSED, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_VERSION, b'\x0d')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_CPU_CAPABILITIES, b'\x00\x00\x00\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NCPUS, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_UNUSED0, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_CACHE_LINESIZE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_SCHED_GEN, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_MEMORY_PRESSURE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_SPIN_COUNT, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_ACTIVE_CPUS, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_PHYSICAL_CPUS, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_LOGICAL_CPUS, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_UNUSED1, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_MEMORY_SIZE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_CPUFAMILY, b'\xec\x5e\x3b\x57')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_KDEBUG_ENABLE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_ATM_DIAGNOSTIC_CONFIG, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_UNUSED2, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_TIME_DATA_START, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NT_TSC_BASE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NT_SCALE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NT_SHIFT, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NT_NS_BASE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_NT_GENERATION, b'\x01')       # someflag seem important 
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_GTOD_GENERATION, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_GTOD_NS_BASE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_GTOD_SEC_BASE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_APPROX_TIME, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_APPROX_TIME_SUPPORTED, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_CONT_TIMEBASE, b'\x00')
    ql.mem.write(COMM_PAGE_START_ADDRESS + COMM_PAGE_BOOTTIME_USEC, b'\x00')


class QlLoaderMACHO(QlLoader):
    # macho x8664 loader 
    def __init__(self, ql, dyld_path=None):
        super(QlLoaderMACHO, self).__init__(ql)
        self.dyld_path      = dyld_path
        self.ql             = ql

        #FIXME: Demigod needs a better way to handle kext file
        if os.path.isdir(self.ql.argv[0]):
            basename = os.path.basename(self.ql.argv[0])
            self.kext_name = os.path.splitext(basename)[0]
            filename = self.ql.argv
            self.ql._argv = [self.ql.argv[0] + "/Contents/MacOS/" + self.kext_name]
            self.ql._path = self.ql.argv[0]
            self.plist = plistlib.load(open(filename[0] + "/Contents/Info.plist", "rb"))
            if "IOKitPersonalities" in self.plist:
                self.IOKit = True
            else:
                self.IOKit = False
        else:
            self.kext_name = None        
    
    def run(self):
        self.profile        = self.ql.profile
        stack_address      = int(self.profile.get("OS64", "stack_address"), 16)
        stack_size         = int(self.profile.get("OS64", "stack_size"), 16)
        vmmap_trap_address = int(self.profile.get("OS64", "vmmap_trap_address"), 16)
        self.heap_address = int(self.profile.get("OS64", "heap_address"), 16)
        self.heap_size = int(self.profile.get("OS64", "heap_size"), 16)        
        self.stack_address = stack_address
        self.stack_size = stack_size

        if self.ql.code:
            self.ql.mem.map(self.ql.os.entry_point, self.ql.os.code_ram_size, info="[shellcode_stack]")
            self.ql.os.entry_point  = (self.ql.os.entry_point + 0x200000 - 0x1000)
            
            self.ql.mem.write(self.entry_point, self.ql.code)

            self.ql.reg.arch_sp = self.ql.os.entry_point
            return
        
        self.ql.os.macho_task = MachoTask()
        self.ql.os.macho_fs = FileSystem(self.ql)
        self.ql.os.macho_mach_port = MachPort(2187)
        self.ql.os.macho_port_manager = MachPortManager(self.ql, self.ql.os.macho_mach_port)
        self.ql.os.macho_host_server = MachHostServer(self.ql)
        self.ql.os.macho_task_server = MachTaskServer(self.ql)
        
        self.envs = env_dict_to_array(self.env)
        self.apples = self.ql.os.transform_to_relative_path(self.ql.path)
        self.ql.os.heap = QlMemoryHeap(self.ql, self.heap_address, self.heap_address + self.heap_size)

        # FIXME: Not working due to overlarge mapping, need to fix it
        # vm_shared_region_enter(self.ql)

        map_commpage(self.ql)

        self.ql.os.thread_management = QlMachoThreadManagement(self.ql)
        self.ql.os.macho_thread = QlMachoThread(self.ql)
        self.ql.os.thread_management.cur_thread = self.ql.os.macho_thread
        self.ql.os.macho_vmmap_end = vmmap_trap_address
        self.stack_sp = stack_address + stack_size
        self.macho_file     = MachoParser(self.ql, self.ql.path)
        self.is_driver      = (self.macho_file.header.file_type == 0xb)
        self.loading_file   = self.macho_file
        self.slide          = int(self.profile.get("LOADER", "slide"), 16)
        self.dyld_slide     = int(self.profile.get("LOADER", "dyld_slide"), 16)
        self.string_align   = 8
        self.ptr_align      = 8
        self.binary_entry   = 0x0
        self.proc_entry     = 0x0
        self.argvs          = [self.ql.path]
        self.argc           = 1
        self.using_dyld     = False
        self.vm_end_addr    = 0x0
        self.ql.mem.map(self.stack_address, self.stack_size, info="[stack]")
        if self.is_driver:
            self.loadDriver(self.stack_address)
            self.ql.hook_code(hook_kernel_api)
        else:
            self.loadMacho()
        self.stack_address = (int(self.stack_sp))
        self.ql.reg.arch_sp = self.stack_address # self.stack_sp
        self.init_sp = self.ql.reg.arch_sp
        self.ql.os.macho_task.min_offset = page_align_end(self.vm_end_addr, PAGE_SIZE)

    def loadDriver(self, stack_addr, loadbase = -1, argv = [], env = {}):
        self.import_symbols = {}
        PAGE_SIZE = 0x1000
        if loadbase < 0:
            loadbase = 0xffffff7000000000
        self.slide = loadbase
        self.load_address = loadbase
        cmds = self.macho_file.commands
        for cmd in cmds:
            if cmd.cmd_id == LC_SEGMENT_64:
                self.loadSegment64(cmd, False)

        self.kext_size = self.vm_end_addr - loadbase

        kernel_path = os.path.join(self.ql.rootfs, "System/Library/Kernels/kernel.development")
        self.ql.log.info("Parsing kernel:")
        self.kernel = MachoParser(self.ql, kernel_path)

        # Create memory for external static symbol jmp code
        self.static_addr = self.vm_end_addr
        self.static_size = PAGE_SIZE
        self.ql.mem.map(self.static_addr, self.static_size, info="[STATIC]")
        self.vm_end_addr += PAGE_SIZE
        self.ql.log.info("Memory for external static symbol is created at 0x%x with size 0x%x" % (self.static_addr, self.static_size))
        self.static_symbols = {}

        # Load kernel
        self.slide = 0
        self.loading_file = self.kernel
        kern_cmds = self.kernel.commands
        self.kernel_base = None
        for cmd in kern_cmds:
            if cmd.cmd_id == LC_SEGMENT_64:
                if self.kernel_base is None:
                    self.kernel_base = cmd.vm_address
                self.loadSegment64(cmd, False)

        self.ql.log.info("Kernel loaded at 0x%x" % self.kernel_base)

        # Resolve local relocation
        for relocation in self.macho_file.dysymbol_table.locreloc:
            seg = None
            for segment in self.macho_file.segments:
                if relocation.symbolnum in segment.sections_index:
                    seg = segment
                    break
            current_value, = struct.unpack("<Q", self.ql.mem.read(loadbase + relocation.address, 8))
            self.ql.log.debug("Patching relocation (0x%x): from 0x%x, update to segment %s at 0x%x" % (loadbase + relocation.address, current_value, seg.name, loadbase + seg.vm_address))
            self.ql.mem.write(loadbase + relocation.address, struct.pack("<Q", current_value + loadbase ))

        # Resolve dynamic symbols
        kernel_local_symbols_index = self.kernel.dysymbol_table.locsymbol_index
        kernel_local_symbols_num = self.kernel.dysymbol_table.locsymbol_num
        self.kernel_local_symbols_detail = self.kernel.symbol_table.details(kernel_local_symbols_index, kernel_local_symbols_num, self.kernel.string_table)

        for key in self.kernel_local_symbols_detail:
            value = self.kernel_local_symbols_detail[key]
            self.import_symbols[value["n_value"]] = key

        kernel_extrn_symbols_index = self.kernel.dysymbol_table.defext_index
        kernel_extrn_symbols_num = self.kernel.dysymbol_table.defext_num
        self.kernel_extrn_symbols_detail = self.kernel.symbol_table.details(kernel_extrn_symbols_index, kernel_extrn_symbols_num, self.kernel.string_table)

        for key in self.kernel_extrn_symbols_detail:
            value = self.kernel_extrn_symbols_detail[key]
            self.import_symbols[value["n_value"]] = key

        offset = 0
        """
        0:  48 83 ec 08             sub    rsp,0x8
        4:  c7 04 24 af be ad de    mov    DWORD PTR [rsp],0xdeadbeaf
        b:  c7 44 24 04 be ba fe    mov    DWORD PTR [rsp+0x4],0xcafebabe
        12: ca
        13: c3                      ret
        """

        for relocation in self.macho_file.dysymbol_table.extreloc:
            symbol = self.macho_file.symbol_table.symbols[relocation.symbolnum]
            symname = self.macho_file.string_table[symbol.n_strx]
            if relocation.length == 2 and relocation.rtype == 2:
                if symname not in self.static_symbols:
                    if symname in self.kernel_local_symbols_detail:
                        real_addr = self.kernel_local_symbols_detail[symname]["n_value"]
                    elif b"_" + symname in self.kernel_extrn_symbols_detail:
                        # ___MALLOC ftw???
                        real_addr = self.kernel_extrn_symbols_detail[b"_" + symname]["n_value"]
                    elif symname in self.kernel_extrn_symbols_detail:
                        real_addr = self.kernel_extrn_symbols_detail[symname]["n_value"]
                    else:
                        self.ql.log.info("Static symbol %s not found" % symname)
                        continue
                    self.import_symbols[real_addr] = symname
                    lo_addr = real_addr & 0xffffffff
                    hi_addr = (real_addr & 0xffffffff00000000) // 0x100000000
                    jmpcode = b"\x48\x83\xec\x08\xc7\x04\x24" + struct.pack("<I", lo_addr) + b"\xc7\x44\x24\x04" + struct.pack("<I", hi_addr) + b"\xc3"

                    self.ql.mem.write(self.static_addr + offset, jmpcode)

                    self.ql.mem.write(loadbase + relocation.address, struct.pack("<I", self.static_addr + offset - (loadbase + relocation.address + 4)))
                    self.static_symbols[symname] = self.static_addr + offset
                    offset += len(jmpcode)
                else:
                    self.ql.mem.write(loadbase + relocation.address, struct.pack("<I", self.static_symbols[symname] - (loadbase + relocation.address + 4)))

#                 ql.log.info("Patching relocation (0x%x): %s at 0x%x" % (loadbase + relocation.address, symname, self.static_symbols[symname]))
                continue
            if relocation.extern == 0 or relocation.length != 3:
                continue

            if symname in self.kernel_local_symbols_detail:
                # ql.log.debug("Patching relocation (0x%x): %s at 0x%x" % (loadbase + relocation.address, symname, self.kernel_local_symbols_detail[symname]["n_value"]))
                self.ql.mem.write(loadbase + relocation.address, struct.pack("<Q", self.kernel_local_symbols_detail[symname]["n_value"]))
            elif symname in self.kernel_extrn_symbols_detail:
                # ql.log.debug("Patching relocation (0x%x): %s at 0x%x" % (loadbase + relocation.address, symname, self.kernel_extrn_symbols_detail[symname]["n_value"]))
                self.ql.mem.write(loadbase + relocation.address, struct.pack("<Q", self.kernel_extrn_symbols_detail[symname]["n_value"]))
            else:
                self.ql.log.info("Symbol %s not found!" % symname)

        # Update resolved symbols in table
        self.loadbase = loadbase
        index = self.macho_file.dysymbol_table.locsymbol_index
        num = self.macho_file.dysymbol_table.locsymbol_num
        self.kext_local_symbols = self.macho_file.symbol_table.details(index, num, self.macho_file.string_table)

        index = self.macho_file.dysymbol_table.defext_index
        num = self.macho_file.dysymbol_table.defext_num
        self.kext_extern_symbols = self.macho_file.symbol_table.details(index, num, self.macho_file.string_table)

        if self.IOKit is True:
            # Get exported vtables
            self.vtables = {}
            for symbol in self.macho_file.symbol_table.symbols:
                if symbol.n_type == 0xf and "__const".ljust(16, "\x00") == self.macho_file.sections[symbol.n_sect].name:
                    symname = self.macho_file.string_table[symbol.n_strx]
                    self.ql.log.info("Found vtable of %s at 0x%x" % (symname, loadbase + symbol.n_value))
                    self.vtables[symname] = loadbase + symbol.n_value
     
            kext = self.plist["IOKitPersonalities"][self.kext_name]["IOClass"]
            user = self.plist["IOKitPersonalities"][self.kext_name]["IOUserClientClass"]

            self.kext_alloc = None
            self.kext_init = None
            self.kext_attach = None
            self.kext_probe = None
            self.kext_detach = None
            self.kext_start = None

            # No need to detach since we will emulate kext and user together
            self.user_alloc = None
            self.user_initWithTask = None
            self.user_attach = None
            self.user_start = None

            for relocation in self.macho_file.dysymbol_table.extreloc:
                symbol = self.macho_file.symbol_table.symbols[relocation.symbolnum]
                symname = self.macho_file.string_table[symbol.n_strx]
                if b"externalMethod" in symname:
                    current_value, = struct.unpack("<Q", self.ql.mem.read(loadbase + relocation.address, 8))
                    print(symname, hex(relocation.address), hex(current_value))

            for symname in self.vtables:
                # TODO: Use IDA Pro to dump offset of methods of IOService and IOUserClient objects
                if symname.decode().endswith(str(len(kext)) + kext + "9MetaClassE"):
                    self.kext_alloc, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x98, 8))
                elif symname.decode().endswith(str(len(kext)) + kext):
                    self.kext_init, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x258, 8))
                    self.kext_attach, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x680, 8))
                    self.kext_probe, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x5c8, 8))
                    self.kext_detach, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x688, 8))
                    self.kext_start, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x5d0, 8))
                elif symname.decode().endswith(str(len(user)) + user + "9MetaClassE"):
                    self.user_alloc, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x98, 8))
                elif symname.decode().endswith(str(len(user)) + user):
                    self.user_initWithTask, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x8f0, 8))
                    self.user_attach, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x680, 8))
                    self.user_start, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x5d0, 8))
                    self.user_externalMethod, = struct.unpack("<Q", self.ql.mem.read(self.vtables[symname] + 0x860, 8))
        else:
#             from pprint import pprint
#             pprint(kext_local_symbols)
            if b"__realmain" in self.kext_local_symbols:
                realmain = loadbase + self.kext_local_symbols[b"__realmain"]["n_value"]
                current_value, = struct.unpack("<Q", self.ql.mem.read(realmain, 8))
                self.ql.log.info("Found entry point: 0x%x" % (current_value))
                self.kext_start = current_value
            else:
                self.ql.log.info("Entry point not found")
                self.kext_start = None

            if b"__antimain" in self.kext_local_symbols:
                antimain = loadbase + self.kext_local_symbols[b"__antimain"]["n_value"]
                current_value, = struct.unpack("<Q", self.ql.mem.read(antimain, 8))
                self.ql.log.info("Found exit point: 0x%x" % (current_value))
                self.kext_stop = current_value
            else:
                self.ql.log.info("Exit point not found")
                self.kext_stop = None

            self.slide = loadbase 

    def loadMacho(self, depth=0, isdyld=False):
        mmap_address   = int(self.profile.get("OS64", "mmap_address"), 16)

        # MAX load depth 
        if depth > 5:
            return

        # three pass 
        # 1: unixthread, uuid, code signature
        # 2: segment
        # 3: dyld
        for pass_count in range(1, 4):

            if isdyld:
                cmds = self.dyld_file.commands
            else:
                cmds = self.macho_file.commands

            for cmd in cmds:
                if pass_count == 1:
                    if cmd.cmd_id == LC_UNIXTHREAD:
                        self.loadUnixThread(cmd, isdyld)

                    if cmd.cmd_id == LC_UUID:
                        self.loadUuid()

                    if cmd.cmd_id == LC_CODE_SIGNATURE:
                        self.loadCodeSignature()

                    if cmd.cmd_id == LC_MAIN:
                        self.loadMain(cmd)

                if pass_count == 2:
                    if cmd.cmd_id == LC_SEGMENT:
                        pass

                    if cmd.cmd_id == LC_SEGMENT_64:
                        self.loadSegment64(cmd, isdyld)

                if pass_count == 3:
                    if cmd.cmd_id == LC_LOAD_DYLINKER:
                        self.loadDylinker(cmd)
                        self.using_dyld = True
                        if not isdyld:
                            if not self.dyld_path:
                                raise QlErrorMACHOFormat("Error No Dyld path")
                            self.dyld_path =  os.path.join(self.ql.rootfs + self.dyld_path)
                            self.dyld_file = MachoParser(self.ql, self.dyld_path)
                            self.loading_file = self.dyld_file
                            self.proc_entry = self.loadMacho(depth + 1, True)
                            self.loading_file = self.macho_file
                            self.using_dyld = True

        if depth == 0:
            self.mmap_address = mmap_address
            self.stack_sp = self.loadStack()
            if self.using_dyld:
                self.ql.log.info("ProcEntry: {}".format(hex(self.proc_entry)))
                self.entry_point = self.proc_entry + self.dyld_slide
                self.ql.log.info("Dyld entry point: {}".format(hex(self.entry_point)))
            else:
                self.entry_point = self.proc_entry + self.slide
            self.ql.log.info("Binary Entry Point: 0x{:X}".format(self.binary_entry))
            self.macho_entry = self.binary_entry + self.slide
            self.load_address = self.macho_entry

        # load_commpage not wroking with ARM64, yet
        if  self.ql.archtype== QL_ARCH.X8664:
            load_commpage(self.ql)

        return self.proc_entry
        
    def loadSegment64(self, cmd, isdyld):
        PAGE_SIZE = 0x1000
        if isdyld:
            slide = self.dyld_slide
        else:
            slide = self.slide
        vaddr_start = cmd.vm_address + slide
        vaddr_end = cmd.vm_address + cmd.vm_size + slide 
        seg_size = cmd.vm_size
        seg_name = cmd.segment_name
        seg_data = bytes(self.loading_file.get_segment(seg_name).content)

        if seg_size == 0:
            return -1

        if seg_name[:10] == "__PAGEZERO":
            self.ql.log.debug("Now loading {}, VM[{}:{}] for pagezero actually it only got a page size".format(seg_name, hex(vaddr_start), hex(vaddr_end)))
            self.ql.mem.map(vaddr_start, PAGE_SIZE, info="[__PAGEZERO]")
            self.ql.mem.write(vaddr_start, b'\x00' * PAGE_SIZE)
            if self.vm_end_addr < vaddr_end:
                self.vm_end_addr = vaddr_end
        else:
            if vaddr_end % PAGE_SIZE != 0:
                vaddr_end = ((vaddr_end // PAGE_SIZE) + 1) * PAGE_SIZE
                seg_size = vaddr_end - vaddr_start
                seg_data = seg_data.ljust(seg_size, b'\0')

            self.ql.log.debug("Now loading {}, VM[{}:{}]".format(seg_name, hex(vaddr_start), hex(vaddr_end)))
            self.ql.mem.map(vaddr_start, seg_size,  info="[loadSegment64]")
            self.ql.mem.write(vaddr_start, seg_data)
            if self.vm_end_addr < vaddr_end:
                self.vm_end_addr = vaddr_end

        return vaddr_start
    
    def loadUnixThread(self, cmd, isdyld):
        if not isdyld:
            self.binary_entry = cmd.entry
 
        self.proc_entry = cmd.entry
        self.ql.log.debug("Binary Thread Entry: {}".format(hex(cmd.entry)))


    def loadUuid(self):
        # todo: WIP
        pass

    def loadCodeSignature(self):
        # todo: WIP
        pass
    
    def loadMain(self, cmd, isdyld=False):
        if self.macho_file.page_zero_size:
            if not isdyld:
                self.binary_entry = cmd.entry_offset + self.macho_file.page_zero_size
            self.proc_entry = cmd.entry_offset + self.macho_file.page_zero_size

    def loadDylinker(self, cmd):
        self.dyld_path = cmd.name

    def make_string(self, argvs, envs, apple_str):
        result = bytes()
        for item in apple_str:
            b = bytes(item, encoding='utf8') + b'\x00'
            result = b +result
        for item in envs:
            b = bytes(item, encoding='utf8') + b'\x00'
            result = b +result 
        for item in argvs:
            b = bytes(item, encoding='utf8') + b'\x00'
            result = b +result
        return result 

    # TODO: add size check
    def loadStack(self):

        argvs_ptr = []
        envs_ptr = []
        apple_ptr = []

        all_str = self.make_string(self.argvs, self.envs, self.apples)
        self.push_stack_string(all_str)
        ptr = self.stack_sp

        for item in self.argvs[::-1]:
            argvs_ptr.append(ptr)  # need pack and tostring
            self.ql.log.debug('add argvs ptr {}'.format(hex(ptr)))
            ptr += len(item) + 1
        
        for item in self.envs[::-1]:
            envs_ptr.append(ptr)
            self.ql.log.debug('add envs ptr {}'.format(hex(ptr)))
            ptr += len(item) + 1

        for item in self.apples[::-1]:
            apple_ptr.append(ptr)
            self.ql.log.debug('add apple ptr {}'.format(hex(ptr)))
            ptr += len(item) + 1

        ptr = self.stack_sp
        self.push_stack_addr(0x0)
        ptr -= 4
        
        for item in apple_ptr:
            self.push_stack_addr(item)
            ptr -= 4
        
        self.push_stack_addr(0x0)
        ptr -= 4
        
        for item in envs_ptr:
            ptr -= 4
            self.push_stack_addr(item)

        self.push_stack_addr(0x0)
        ptr -= 4
        
        for item in argvs_ptr:
            ptr -= 4
            self.push_stack_addr(item)
            self.ql.log.debug("SP 0x%x, content 0x%x" % (self.stack_sp, item))
        argvs_ptr_ptr = ptr 

        self.push_stack_addr(self.argc)
        ptr -= 4
        self.ql.log.debug("SP 0x%x, content 0x%x" % (self.stack_sp, self.argc))
       
        if self.using_dyld:
            ptr -= 4
            #ql.log.info("Binary Dynamic Entry Point: {:X}".format(self.binary_entry))
            self.push_stack_addr(self.macho_file.header_address)
            # self.push_stack_addr(self.binary_entry)

        return self.stack_sp

    def push_stack_string(self, data):
        align = self.string_align
        length = len(data)
        
        if length % align != 0:
            for i in range(align - (length % align)):
                data += b'\x00' 
            length = len(data)
        
        self.stack_sp -= length
        self.ql.mem.write(self.stack_sp, data)
        self.ql.log.debug("SP {} write data len {}".format(hex(self.stack_sp), length))
        
        return self.stack_sp
    
    def push_stack_addr(self, data):
        align = self.ptr_align
        
        if data == 0:
            content = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        else:
            content = struct.pack('<Q', data)

        if len(content) != align:
            self.ql.log.info('stack align error')
            return 
        
        self.stack_sp -= align
        self.ql.mem.write(self.stack_sp, content)

        return self.stack_sp



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from functools import partial

from qiling.const import *

CODE_END = True 


def dump_regs(ql):

    if ql.archtype == QL_ARCH.MIPS:

        _reg_order = (
                "gp", "at", "v0", "v1",
                "a0", "a1", "a2", "a3",
                "t0", "t1", "t2", "t3",
                "t4", "t5", "t6", "t7",
                "t8", "t9", "sp", "s8",
                "s0", "s1", "s2", "s3",
                "s4", "s5", "s6", "s7",
                "ra", "k0", "k1", "pc",
                )

    elif ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB):

        _reg_order = (
                "r0", "r1", "r2", "r3",
                "r4", "r5", "r6", "r7",
                "r8", "r9", "r10", "r11",
                "r12", "sp", "lr", "pc",
                )

    return {reg_name: getattr(ql.reg, reg_name) for reg_name in _reg_order}


def get_arm_flags(bits):
    def _get_mode(bits):
        return {
                0b10000: "User",
                0b10001: "FIQ",
                0b10010: "IRQ",
                0b10011: "Supervisor",
                0b10110: "Monitor",
                0b10111: "Abort",
                0b11010: "Hypervisor",
                0b11011: "Undefined",
                0b11111: "System",
                }.get(bits & 0x00001f)

    return {
            "mode":     _get_mode(bits),
            "thumb":    bits & 0x00000020 != 0,
            "fiq":      bits & 0x00000040 != 0,
            "irq":      bits & 0x00000080 != 0,
            "neg":      bits & 0x80000000 != 0,
            "zero":     bits & 0x40000000 != 0,
            "carry":    bits & 0x20000000 != 0,
            "overflow": bits & 0x10000000 != 0,
            }


# parse unsigned integer from string 
def parse_int(s):
    return int(s, 16) if s.startswith("0x") else int(s)


# check wether negative value or not
def is_negative(i):
    return i & (1 << 31)


# convert valu to signed
def signed_val(i):
    val = i

    if is_negative(val):
        val -= 1 << 32

    return val


# handle braches and jumps so we can set berakpoint properly
def handle_bnj(ql, cur_addr):
    return {
            QL_ARCH.MIPS     : handle_bnj_mips,
            QL_ARCH.ARM      : handle_bnj_arm,
            QL_ARCH.ARM_THUMB: handle_bnj_arm,
            }.get(ql.archtype)(ql, cur_addr)


def get_cpsr(bits):
    return (
            bits & 0x10000000 != 0, # V, overflow flag
            bits & 0x20000000 != 0, # C, carry flag
            bits & 0x40000000 != 0, # Z, zero flag
            bits & 0x80000000 != 0, # N, sign flag
            )


def is_thumb(bits):
    return bits & 0x00000020 != 0


def disasm(ql, address):
    md = ql.create_disassembler()
    return next(md.disasm(_read_inst(ql, address), address))


def _read_inst(ql, addr):

    if ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB):
        result = ql.mem.read(addr, 4) # default option for arm instruction
        if is_thumb(ql.reg.cpsr):

            first_two = ql.unpack16(ql.mem.read(addr, 2))
            result = ql.pack16(first_two)

            if any([
                first_two & 0xf000 == 0xf000,
                first_two & 0xf800 == 0xf800,
                first_two & 0xe800 == 0xe800,
                ]):

                latter_two = ql.unpack16(ql.mem.read(addr+2, 2))
                result += ql.pack16(latter_two)

    elif ql.archtype == QL_ARCH.MIPS:
        result = ql.mem.read(addr, 4)

    return result


def handle_bnj_arm(ql, cur_addr):

    def _read_reg_val(regs, _reg):
        return getattr(ql.reg, _reg.replace("ip", "r12"))

    def regdst_eq_pc(op_str):
        return op_str.partition(", ")[0] == "pc"


    read_inst = partial(_read_inst, ql)
    read_reg_val = partial(_read_reg_val, ql.reg)

    ARM_INST_SIZE = 4
    ARM_THUMB_INST_SIZE = 2

    line = disasm(ql, cur_addr)
    ret_addr = cur_addr + line.size

    if line.mnemonic == "udf": # indicates program exited
        return CODE_END

    jump_table = {
            # unconditional branch
            "b"    : (lambda *_: True),
            "bl"   : (lambda *_: True),
            "bx"   : (lambda *_: True),
            "blx"  : (lambda *_: True),
            "b.w"  : (lambda *_: True),

            # branch on equal, Z == 1
            "beq"  : (lambda V, C, Z, N: Z == 1),
            "bxeq" : (lambda V, C, Z, N: Z == 1),
            "beq.w": (lambda V, C, Z, N: Z == 1),

            # branch on not equal, Z == 0
            "bne"  : (lambda V, C, Z, N: Z == 0),
            "bxne" : (lambda V, C, Z, N: Z == 0),
            "bne.w": (lambda V, C, Z, N: Z == 0),

            # branch on signed greater than, Z == 0 and N == V
            "bgt"  : (lambda V, C, Z, N: (Z == 0 and N == V)),
            "bgt.w": (lambda V, C, Z, N: (Z == 0 and N == V)),

            # branch on signed less than, N != V
            "blt"  : (lambda V, C, Z, N: N != V),

            # branch on signed greater than or equal, N == V
            "bge"  : (lambda V, C, Z, N: N == V),

            # branch on signed less than or queal
            "ble"  : (lambda V, C, Z, N: Z == 1 or N != V),

            # branch on unsigned higher or same (or carry set), C == 1
            "bhs"  : (lambda V, C, Z, N: C == 1),
            "bcs"  : (lambda V, C, Z, N: C == 1),

            # branch on unsigned lower (or carry clear), C == 0
            "bcc"  : (lambda V, C, Z, N: C == 0),
            "blo"  : (lambda V, C, Z, N: C == 0),
            "bxlo" : (lambda V, C, Z, N: C == 0),
            "blo.w": (lambda V, C, Z, N: C == 0),

            # branch on negative or minus, N == 1
            "bmi"  : (lambda V, C, Z, N: N == 1),

            # branch on positive or plus, N == 0
            "bpl"  : (lambda V, C, Z, N: N == 0),

            # branch on signed overflow
            "bvs"  : (lambda V, C, Z, N: V == 1),

            # branch on no signed overflow
            "bvc"  : (lambda V, C, Z, N: V == 0),

            # branch on unsigned higher
            "bhi"  : (lambda V, C, Z, N: (Z == 0 and C == 1)),
            "bxhi" : (lambda V, C, Z, N: (Z == 0 and C == 1)),
            "bhi.w": (lambda V, C, Z, N: (Z == 0 and C == 1)),

            # branch on unsigned lower
            "bls"  : (lambda V, C, Z, N: (C == 0 or Z == 1)),
            "bls.w": (lambda V, C, Z, N: (C == 0 or Z == 1)),
            }

    cb_table = {
            # branch on equal to zero
            "cbz" : (lambda r: r == 0),

            # branch on not equal to zero
            "cbnz": (lambda r: r != 0),
            }

    to_jump = False
    if line.mnemonic in jump_table:
        to_jump = jump_table.get(line.mnemonic)(*get_cpsr(ql.reg.cpsr))

    elif line.mnemonic in cb_table:
        to_jump = cb_table.get(line.mnemonic)(read_reg_val(line.op_str.split(", ")[0]))

    if to_jump:
        if '#' in line.op_str:
            ret_addr = parse_int(line.op_str.split('#')[-1])
        else:
            ret_addr = read_reg_val(line.op_str)

            if regdst_eq_pc(line.op_str):
                next_addr = cur_addr + line.size
                n2_addr = next_addr + len(read_inst(next_addr))
                ret_addr += len(read_inst(n2_addr)) + len(read_inst(next_addr))

    elif line.mnemonic.startswith("it"):
        # handle IT block here

        cond_met = {
                "eq": lambda V, C, Z, N: (Z == 1),
                "ne": lambda V, C, Z, N: (Z == 0),
                "ge": lambda V, C, Z, N: (N == V),
                "hs": lambda V, C, Z, N: (C == 1),
                "lo": lambda V, C, Z, N: (C == 0),
                "mi": lambda V, C, Z, N: (N == 1),
                "pl": lambda V, C, Z, N: (N == 0),
                "ls": lambda V, C, Z, N: (C == 0 or Z == 1),
                "le": lambda V, C, Z, N: (Z == 1 or N != V),
                "hi": lambda V, C, Z, N: (Z == 0 and C == 1),
                }.get(line.op_str)(*get_cpsr(ql.reg.cpsr))
        
        it_block_range = [each_char for each_char in line.mnemonic[1:]]

        next_addr = cur_addr + ARM_THUMB_INST_SIZE
        for each in it_block_range:
            _inst = read_inst(next_addr)
            n2_addr = handle_bnj_arm(ql, next_addr)

            if (cond_met and each == "t") or (not cond_met and each == "e"):
                if n2_addr != (next_addr+len(_inst)): # branch detected
                    break

            next_addr += len(_inst)

        ret_addr = next_addr

        return ret_addr

    elif line.mnemonic in ("ldr",):
        if regdst_eq_pc(line.op_str):
            _pc, _, rn_offset = line.op_str.partition(", ")

            if "]" in rn_offset.split(", ")[1]: # pre-indexed immediate
                _, r, imm = line.op_str.replace("[", "").replace("]", "").replace("!", "").replace("#", "").split(", ")
                ret_addr = ql.unpack32(ql.mem.read(parse_int(imm) + read_reg_val(r), 4))

            else: # post-indexed immediate
                # FIXME: weired behavior, immediate here does not apply
                _, r, imm = line.op_str.replace("[", "").replace("]", "").replace("!", "").replace("#", "").split(", ")
                ret_addr = ql.unpack32(ql.mem.read(read_reg_val(r), 4))

    elif line.mnemonic in ("addls", "addne", "add") and regdst_eq_pc(line.op_str):
        V, C, Z, N = get_cpsr(ql.reg.cpsr)

        if line.mnemonic == "addls" and (C == 0 or Z == 1):
            r0, r1, r2, imm = line.op_str.split(", ")
            # program counter is awalys 8 bytes ahead , when it comes with pc need to add extra 8 bytes
            ret_addr = 8 + read_reg_val(r1) + read_reg_val(r2) * 4

        elif line.mnemonic == "addne" and Z == 0:
            r0, r1, r2, *rest = line.op_str.split(", ")
            ret_addr = 8 + read_reg_val(r1) + (read_reg_val(r2) * 4 if rest else read_reg_val(r2))

        elif line.mnemonic == "add":
            r0, r1, r2 = line.op_str.split(", ")
            ret_addr = 8 + sum(map(read_reg_val, [r1, r2]))

    elif line.mnemonic in ("tbh", "tbb"):

        cur_addr += ARM_INST_SIZE

        if line.mnemonic == "tbh":
            r0, r1, _ = line.op_str.strip("[").strip("]").split(", ")
            r1 = read_reg_val(r1) * 2

        elif line.mnemonic == "tbb":
            r0, r1 = line.op_str.strip("[").strip("]").split(", ")
            r1 = read_reg_val(r1)

        to_add = int.from_bytes(ql.mem.read(cur_addr+r1, 2 if line.mnemonic == "tbh" else 1), byteorder="little") * 2
        ret_addr = cur_addr + to_add

    elif line.mnemonic.startswith("pop") and "pc" in line.op_str:

        ret_addr = ql.stack_read(line.op_str.strip("{").strip("}").split(", ").index("pc") * 4)
        if not { # step to next instruction if cond does not meet
                "pop"  : lambda *_: True,
                "pop.w": lambda *_: True,
                "popeq": lambda V, C, Z, N: (Z == 1),
                "popne": lambda V, C, Z, N: (Z == 0),
                "pophi": lambda V, C, Z, N: (C == 1),
                "popge": lambda V, C, Z, N: (N == V),
                "poplt": lambda V, C, Z, N: (N != V),
                }.get(line.mnemonic)(*get_cpsr(ql.reg.cpsr)):

            ret_addr = cur_addr + ARM_INST_SIZE

    elif line.mnemonic == "sub" and regdst_eq_pc(line.op_str):
        _, r, imm = line.op_str.split(", ")
        ret_addr = read_reg_val(r) - parse_int(imm.strip("#"))

    elif line.mnemonic == "mov" and regdst_eq_pc(line.op_str):
        _, r = line.op_str.split(", ")
        ret_addr = read_reg_val(r)

    if ret_addr & 1:
        ret_addr -= 1

    return ret_addr


def handle_bnj_mips(ql, cur_addr):
    MIPS_INST_SIZE = 4

    def _read_reg(regs, _reg):
        return signed_val(getattr(regs, _reg.strip('$').replace("fp", "s8")))

    read_reg_val = partial(_read_reg, ql.reg)

    line = disasm(ql, cur_addr)

    # default breakpoint address if no jumps and branches here
    ret_addr = cur_addr + MIPS_INST_SIZE

    if line.mnemonic.startswith('j') or line.mnemonic.startswith('b'):
        if line.mnemonic == "break": # indicates program extied
            return CODE_END

        # make sure at least delay slot executed
        ret_addr += MIPS_INST_SIZE

        # get registers or memory address from op_str
        targets = [
                read_reg_val(each)
                if '$' in each else parse_int(each)
                for each in line.op_str.split(", ")
                ]

        to_jump = {
                "j"       : (lambda _: True),             # uncontitional jump
                "jr"      : (lambda _: True),             # uncontitional jump
                "jal"     : (lambda _: True),             # uncontitional jump
                "jalr"    : (lambda _: True),             # uncontitional jump
                "b"       : (lambda _: True),             # unconditional branch
                "bl"      : (lambda _: True),             # unconditional branch
                "bal"     : (lambda _: True),             # unconditional branch
                "beq"     : (lambda r0, r1, _: r0 == r1), # branch on equal
                "bne"     : (lambda r0, r1, _: r0 != r1), # branch on not equal
                "blt"     : (lambda r0, r1, _: r0 < r1),  # branch on r0 less than r1
                "bgt"     : (lambda r0, r1, _: r0 > r1),  # branch on r0 greater than r1
                "ble"     : (lambda r0, r1, _: r0 <= r1), # brach on r0 less than or equal to r1
                "bge"     : (lambda r0, r1, _: r0 >= r1), # branch on r0 greater than or equal to r1
                "beqz"    : (lambda r, _: r == 0),        # branch on equal to zero
                "bnez"    : (lambda r, _: r != 0),        # branch on not equal to zero
                "bgtz"    : (lambda r, _: r > 0),         # branch on greater than zero
                "bltz"    : (lambda r, _: r < 0),         # branch on less than zero
                "bltzal"  : (lambda r, _: r < 0),         # branch on less than zero and link
                "blez"    : (lambda r, _: r <= 0),        # branch on less than or equal to zero
                "bgez"    : (lambda r, _: r >= 0),        # branch on greater than or equal to zero
                "bgezal"  : (lambda r, _: r >= 0),        # branch on greater than or equal to zero and link
                }.get(line.mnemonic)(*targets)

        if to_jump:
            # target address is always the rightmost one
            ret_addr = targets[-1]

    return ret_addr


def diff_snapshot_save(current_state_dicts, prev_states):

    result = {}
    cur_cpu_ctx_set = {(idx, val) for idx, val in enumerate(current_state_dicts["cpu_context"])}

    cur_mem_set = set()
    for region_idx, mem_region_info in current_state_dicts["mem"].items():
        raw_bytes = bytes(mem_region_info[-1])
        cur_mem_set.update({(region_idx, *(mem_region_info[:-1]), raw_bytes)})

    if prev_states is not None:
        cur_cpu_ctx_set -= prev_states["cpu_context"]
        result.update({"cpu_context": cur_cpu_ctx_set})

        if "mem" in prev_states and cur_mem_set != prev_states["mem"]:
            # save changes if its different
            cur_mem_set -= prev_states["mem"]
            result.update({"mem": cur_mem_set})

    else:
        # store everything since its the first snapshot
        result.update({"cpu_context": cur_cpu_ctx_set, "mem": cur_mem_set})

    return result


def diff_snapshot_restore(current_state_dicts, prev_states):

    result = {}
    _cur_cpu_ctx_dict = {idx: val for idx, val in enumerate(current_state_dicts["cpu_context"])}
    _cur_mem_dict = current_state_dicts["mem"]

    last_cpu_ctx_set = prev_states["cpu_context"]
    last_mem_set = prev_states.pop("mem", None)

    # restore diff snapshot from last cpu_context
    _cur_cpu_ctx_dict.update({addr: raw_byte for addr, raw_byte in last_cpu_ctx_set})
    result.update({"cpu_context": bytes(_cur_cpu_ctx_dict.values())})

    # restore diff snapshot from last memory dump
    if last_mem_set is not None and len(last_mem_set) > 0:
        last_mem_dict = {}
        for region_idx, *region_info, raw_bytes in last_mem_set:
            _region = _cur_mem_dict.get(region_idx, None)
            if _region:
                last_mem_dict.update({region_idx: (*_region[:-1], raw_bytes)})
            else:
                last_mem_dict.update({region_idx: (*region_info, raw_bytes)})

        result.update({"mem": last_mem_dict})

    return result


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

# windows sdk data file can download here: 'https://github.com/ohjeongwook/windows_sdk_data.git'

import json
import os

result_dict = {}


def parse_winsdk(dir):
    def PtrDecl_detect(dic):
        param_dict = {}
        if dic['data_type'] in ['PtrDecl', 'ArrayDecl']:
            return PtrDecl_detect(dic['type'])
        elif dic['data_type'] == 'TypeDecl':
            if 'name' in dic.keys():
                param_dict['name'] = dic['name']
            else:
                param_dict['name'] = 'VOID'
            param_dict['type'] = dic['type']
        elif dic['data_type'] == 'FuncDecl':
            param_dict['name'] = dic['func']['name']
            param_dict['type'] = dic['func']['type']['type']
        else:
            print(dic['data_type'])
            exit(0)
        return param_dict

    if dir != '.DS_Store':
        with open("../data/" + dir, 'r') as f:
            winsdk = json.loads(f.read())
            f.close()
    else:
        return

    entry = 'funcdefs'

    for func in winsdk[entry]:     # read function dict
        if 'name' in func.keys():
            func_name = func['name']
        elif func['type']['data_type'] == 'TypeDecl':
            func_name = func['type']['name']
        else:
            func_name = func['type']['type']['name']

        if 'api_locations' in func.keys():
            func_dll = func['api_locations'][0].lower()
        else:
            continue

        func_dict = {}
        param_list = []
        argument = func['arguments']

        for argu in argument:
            param = PtrDecl_detect(argu)
            param_list.append(param)
            func_dict[func_name] = param_list

        if func_dll in result_dict.keys():
            old_func_dict = result_dict[func_dll].copy()
            result_dict[func_dll] = {}
            result_dict[func_dll] = dict(old_func_dict, **func_dict)
        else:
            result_dict[func_dll] = func_dict


def save2json():
    print("begin to save ...")
    for dllname in result_dict:
        result_json = json.dumps(result_dict[dllname], sort_keys=False, indent=4, separators=(',', ': '))
        print(dllname)
        if dllname[0].isupper():
            dllname = '_'+dllname
        fo = open('./windows_sdk/'+dllname.replace('.', '_')+'.json', 'w+')
        fo.write(result_json)
        fo.close()


if __name__ == '__main__':
    dir = '../data'
    result_dict = {}
    for parent, dirnames, filenames in os.walk(dir, followlinks=True):
        for filename in filenames:
            parse_winsdk(filename)
    save2json()



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_set_robust_list(ql, set_robust_list_head_ptr, set_robust_list_head_len, *args, **kw):
    if ql.multithread == True:
        ql.os.thread_management.cur_thread.robust_list_head_ptr = set_robust_list_head_ptr
        ql.os.thread_management.cur_thread.robust_list_head_len = set_robust_list_head_len
    regreturn = 0
    return regreturn


def ql_syscall_futex(ql, futex_uaddr, futex_op, futex_val, futex_timeout, futex_uaddr2, futex_val3):
    FUTEX_WAIT = 0
    FUTEX_WAKE = 1
    FUTEX_FD = 2
    FUTEX_REQUEUE = 3
    FUTEX_CMP_REQUEUE = 4
    FUTEX_WAKE_OP = 5
    FUTEX_LOCK_PI = 6
    FUTEX_UNLOCK_PI = 7
    FUTEX_TRYLOCK_PI = 8
    FUTEX_WAIT_BITSET = 9
    FUTEX_WAKE_BITSET = 10
    FUTEX_WAIT_REQUEUE_PI = 11
    FUTEX_CMP_REQUEUE_PI = 12
    FUTEX_PRIVATE_FLAG = 128

    if futex_op & (FUTEX_PRIVATE_FLAG - 1) == FUTEX_WAIT:
        # def futex_wait_addr(ql, th, arg):
        #     addr, val = arg
        #     if ql.unpack32(ql.mem.read(addr, 4)) != val:
        #         return False
        #     else:
        #         return True
        regreturn = ql.os.futexm.futex_wait(ql, futex_uaddr,
                                            ql.os.thread_management.cur_thread,
                                            futex_val)
        ql.log.debug("futex(%x, %d, %d, %x) = %d" % (futex_uaddr, futex_op, futex_val, futex_timeout, regreturn))
    elif futex_op & (FUTEX_PRIVATE_FLAG - 1) == FUTEX_WAIT_BITSET:
        regreturn = ql.os.futexm.futex_wait(ql, futex_uaddr,
                                            ql.os.thread_management.cur_thread,
                                            futex_val,
                                            futex_val3)
        ql.log.debug("futex(%x, %d, %d, %x, %x, %x) = %d" % (futex_uaddr,
                                                          futex_op, futex_val,
                                                          futex_timeout,
                                                          futex_uaddr2,
                                                          futex_val3,
                                                          regreturn))
    elif futex_op & (FUTEX_PRIVATE_FLAG - 1) == FUTEX_WAKE:
        regreturn = ql.os.futexm.futex_wake(ql, futex_uaddr,ql.os.thread_management.cur_thread, futex_val)
        ql.log.debug("futex(%x, %d, %d) = %d" % (futex_uaddr, futex_op, futex_val, regreturn))
    elif futex_op & (FUTEX_PRIVATE_FLAG - 1) == FUTEX_WAKE_BITSET:
        regreturn = ql.os.futexm.futex_wake(ql, futex_uaddr,ql.os.thread_management.cur_thread, futex_val, futex_val3)
        ql.log.debug("futex(%x, %d, %d) = %d" % (futex_uaddr, futex_op, futex_val, regreturn))
    else:
        ql.log.debug("futex(%x, %d, %d) = ?" % (futex_uaddr, futex_op, futex_val))
        ql.emu_stop()
        #ql.os.thread_management.cur_thread.stop()
        #ql.os.thread_management.cur_thread.stop_event = THREAD_EVENT_EXIT_GROUP_EVENT
        regreturn = 0

    return regreturn



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from .header import *
from .loadcommand import *
from .data import *
from .const import *
from .utils import *
from struct import unpack
from qiling.const import *


class MachoParser:
    
    # arch = "x8664" or "x86" 
    def __init__(self, ql, path, arch= None):
        self.ql = ql
        self.binary_file = self.readFile(path)
        self.raw_data = self.binary_file
        self.archtype = ql.archtype
        self.parseFile()
        self.page_zero_size = 0
        self.header_address = 0x0
        for seg in self.segments:
            # find page zero
            if seg.vm_address == 0 and seg.file_size == 0:
                ql.log.info("PageZero Size: {:X}".format(seg.vm_size))
                self.page_zero_size = seg.vm_size
                self.header_address = seg.vm_size

    @staticmethod
    def readFile(path):
        with open(path, "rb") as f:
            return f.read()

    def parseFile(self):
        if not self.binary_file:
            return 
        
        if not self.parseHeader():
            return

        if not self.parseLoadCommand():
            return 

        if not self.parseData():
            return 
        

    def parseHeader(self):
        self.magic = self.getMagic(self.binary_file)
        
        if self.magic in MAGIC_64:
            self.ql.log.debug("Got a 64bit Header ")
            self.header = BinaryHeader(self.binary_file)

        #elif self.magic in MAGIC_X86:
        #    # x86
        #    ql.log.debug("Got a x86 Header") 
        #    self.header = BinaryHeader(self.binary_file)

        elif self.magic in MAGIC_FAT:
            # fat 
            self.ql.log.debug("Got a fat header")
            fat = FatHeader(self.binary_file)
            file_info = fat.getBinary(self.archtype)
            self.binary_file = self.binary_file[file_info.offset : file_info.offset + file_info.size]
            self.header = BinaryHeader(self.binary_file)
        else:
            self.ql.log.info("unknow header!")
            return False
        
        if not self.header:
            self.ql.log.info("parse header error")
            return False 

        return True

    def parseLoadCommand(self):
        self.ql.log.debug("Parse LoadCommand")
        if not self.header.lc_num or not self.header.lc_size or not self.header.header_size:
            return False

        FR = FileReader(self.binary_file)
        FR.setOffset(self.header.header_size)
        self.lc_raw = FR.read(self.header.lc_size)
        self.commands = []
        offset = 0

        for i in range(self.header.lc_num):

            if self.header.lc_size >= 8:
                lc = LoadCommand(self.lc_raw[offset:])
            else:
                self.ql.log.info("cmd size overflow")
                return False 

            if self.header.lc_size >= offset + lc.cmd_size:
                complete_cmd = lc.get_complete()
                pass
            else:
                self.ql.log.info("cmd size overflow")
                return False
            
            self.commands.append(complete_cmd)
            
            offset += lc.cmd_size
        
        return True


    def parseData(self):
        self.segments = []      
        self.sections = [None]
        for command in self.commands:
            if command.cmd_id == LC_SEGMENT_64:
                tmp = Segment(command, self.binary_file)
                tmp.sections_index += range(len(self.sections), len(self.sections) + len(tmp.sections))
                self.segments.append(tmp)
                for section in tmp.sections:
                    self.sections.append(section)
            elif command.cmd_id == LC_SEGMENT:
                tmp = Segment(command, self.binary_file)
                self.segments.append(tmp)
                for section in tmp.sections:
                    self.sections.append(section)
            elif command.cmd_id == LC_FUNCTION_STARTS:
                self.function_starts = FunctionStarts(command, self.binary_file)
            elif command.cmd_id == LC_SYMTAB:
                self.symbol_table = SymbolTable(command, self.binary_file)
                self.string_table = StringTable(command, self.binary_file)
            elif command.cmd_id == LC_DATA_IN_CODE:
                self.data_in_code = DataInCode(command, self.binary_file)
            elif command.cmd_id == LC_CODE_SIGNATURE:
                self.code_signature = CodeSignature(command, self.binary_file)
            elif command.cmd_id == LC_SEGMENT_SPLIT_INFO:
                self.seg_split_info = SegmentSplitInfo(command, self.binary_file)
            elif command.cmd_id == LC_DYSYMTAB:
                self.dysymbol_table = DySymbolTable(command, self.binary_file)
        return True
    
    @staticmethod
    def getMagic(binary):
        return unpack("<L", binary[:4])[0]

    def get_segment(self, name):
        for seg in self.segments:
            if seg.name == name:
                return seg
        return None




#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time

from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *

dllname = 'kernel32_dll'

# LPVOID VirtualAlloc(
#   LPVOID lpAddress,
#   SIZE_T dwSize,
#   DWORD  flAllocationType,
#   DWORD  flProtect
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_VirtualAlloc(ql, address, params):
    dwSize = params["dwSize"]
    addr = ql.os.heap.alloc(dwSize)
    return addr


# BOOL VirtualFree(
#   LPVOID lpAddress,
#   SIZE_T dwSize,
#   DWORD  dwFreeType
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_VirtualFree(ql, address, params):
    lpAddress = params["lpAddress"]
    ql.os.heap.free(lpAddress)
    return 1


# BOOL VirtualProtect(
#  LPVOID lpAddress,
#  SIZE_T dwSize,
#  DWORD  flNewProtect,
#  PDWORD lpflOldProtect
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'SIZE_T': 'UINT', 'DWORD': 'UINT'})
def hook_VirtualProtect(ql, address, params):
    return 1


# SIZE_T VirtualQuery(
#  LPCVOID                   lpAddress,
#  PMEMORY_BASIC_INFORMATION lpBuffer,
#  SIZE_T                    dwLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'SIZE_T': 'UINT', 'DWORD': 'UINT'})
def hook_VirtualQuery(ql, address, params):
    """
    typedef struct _MEMORY_BASIC_INFORMATION {
      PVOID  BaseAddress;
      PVOID  AllocationBase;
      DWORD  AllocationProtect;
      SIZE_T RegionSize;
      DWORD  State;
      DWORD  Protect;
      DWORD  Type;
    } MEMORY_BASIC_INFORMATION, *PMEMORY_BASIC_INFORMATION;
    """
    # find chunk,
    base = None
    size = None
    for chunk in ql.os.heap.chunks:
        if chunk.address <= params['lpAddress'] < chunk.address + chunk.size:
            base = chunk.address
            size = chunk.size

    if not base and not size:
        # Page not found
        # printable = sorted(['0x%x-0x%x' % (chunk.address, chunk.address+chunk.size) for chunk in ql.os.heap.chunks])
        # ql.log.debug('Could not find memory chunk containing address 0x%x in %s' % (params['lpAddress'],
        # printable))
        ql.os.last_error = ERROR_INVALID_PARAMETER
        return 0

    mbi = params['lpBuffer']
    ql.mem.write(mbi, base.to_bytes(length=ql.pointersize, byteorder='little'))
    ql.mem.write(mbi + ql.pointersize * 1, base.to_bytes(length=ql.pointersize, byteorder='little'))
    ql.mem.write(mbi + ql.pointersize * 2,
                 (0x40).to_bytes(length=ql.pointersize, byteorder='little'))  # 0x40 = EXECUTE_READ_WRITE
    ql.mem.write(mbi + ql.pointersize * 3, size.to_bytes(length=ql.pointersize, byteorder='little'))
    ql.mem.write(mbi + ql.pointersize * 4,
                 (0x1000).to_bytes(length=ql.pointersize, byteorder='little'))  # 0x1000 == MEM_COMMIT
    ql.mem.write(mbi + ql.pointersize * 5,
                 (0x40).to_bytes(length=ql.pointersize, byteorder='little'))  # 0x40 = EXECUTE_READ_WRITE
    ql.mem.write(mbi + ql.pointersize * 6,
                 (0x40000).to_bytes(length=ql.pointersize, byteorder='little'))  # 0x40000 = MEM_MAPPED

    return ql.pointersize * 7



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

"""
This module is intended for general purpose functions that are only used in qiling.arch
"""

from unicorn import UcError, UC_ERR_READ_UNMAPPED, UC_ERR_FETCH_UNMAPPED
from keystone import *
from capstone import *

from qiling.const import QL_ARCH, QL_ARCH_ALL, QL_ENDIAN, QL_OS, QL_OS_ALL, QL_OUTPUT, QL_DEBUGGER, QL_ARCH_32BIT, QL_ARCH_64BIT, QL_ARCH_16BIT
from qiling.exception import *


def ql_create_disassembler(archtype, archendian, reg_cpsr=None):
    if archtype == QL_ARCH.ARM:  # QL_ARM
        mode = CS_MODE_ARM
        if archendian == QL_ENDIAN.EB:
            # TODO: Test for big endian.
            reg_cpsr_v = 0b100000
            # reg_cpsr_v = 0b000000
        else:
            reg_cpsr_v = 0b100000

        if reg_cpsr & reg_cpsr_v != 0:
            mode = CS_MODE_THUMB

        if archendian == QL_ENDIAN.EB:
            md = Cs(CS_ARCH_ARM, mode)
            # md = Cs(CS_ARCH_ARM, mode + CS_MODE_BIG_ENDIAN)
        else:
            md = Cs(CS_ARCH_ARM, mode)

    elif archtype == QL_ARCH.ARM_THUMB:
        md = Cs(CS_ARCH_ARM, CS_MODE_THUMB)

    elif archtype == QL_ARCH.X86:  # QL_X86
        md = Cs(CS_ARCH_X86, CS_MODE_32)

    elif archtype == QL_ARCH.X8664:  # QL_X86_64
        md = Cs(CS_ARCH_X86, CS_MODE_64)

    elif archtype == QL_ARCH.ARM64:  # QL_ARM64
        md = Cs(CS_ARCH_ARM64, CS_MODE_ARM)

    elif archtype == QL_ARCH.A8086:  # QL_A8086
        md = Cs(CS_ARCH_X86, CS_MODE_16)

    elif archtype == QL_ARCH.MIPS:  # QL_MIPS32
        if archendian == QL_ENDIAN.EB:
            md = Cs(CS_ARCH_MIPS, CS_MODE_MIPS32 + CS_MODE_BIG_ENDIAN)
        else:
            md = Cs(CS_ARCH_MIPS, CS_MODE_MIPS32 + CS_MODE_LITTLE_ENDIAN)

    else:
        raise QlErrorArch("Unknown arch defined in utils.py (debug output mode)")

    return md

def ql_create_assembler(archtype, archendian, reg_cpsr=None):
    if archtype == QL_ARCH.ARM:  # QL_ARM
        mode = KS_MODE_ARM
        if archendian == QL_ENDIAN.EB:
            # TODO: Test for big endian.
            reg_cpsr_v = 0b100000
            # reg_cpsr_v = 0b000000
        else:
            reg_cpsr_v = 0b100000

        if reg_cpsr & reg_cpsr_v != 0:
            mode = KS_MODE_THUMB

        if archendian == QL_ENDIAN.EB:
            ks = Ks(KS_ARCH_ARM, mode)
            # md = Cs(CS_ARCH_ARM, mode + CS_MODE_BIG_ENDIAN)
        else:
            ks = Ks(KS_ARCH_ARM, mode)

    elif archtype == QL_ARCH.ARM_THUMB:
        ks = Ks(KS_ARCH_ARM, KS_MODE_THUMB)

    elif archtype == QL_ARCH.X86:  # QL_X86
        ks = Ks(KS_ARCH_X86, KS_MODE_32)

    elif archtype == QL_ARCH.X8664:  # QL_X86_64
        ks = Ks(KS_ARCH_X86, KS_MODE_64)

    elif archtype == QL_ARCH.ARM64:  # QL_ARM64
        ks = Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN)

    elif archtype == QL_ARCH.A8086:  # QL_A8086
        ks = Ks(KS_ARCH_X86, KS_MODE_16)

    elif archtype == QL_ARCH.MIPS:  # QL_MIPS32
        if archendian == QL_ENDIAN.EB:
            ks = Ks(KS_ARCH_MIPS, KS_MODE_MIPS32 + KS_MODE_BIG_ENDIAN)
        else:
            ks = Ks(KS_ARCH_MIPS, KS_MODE_MIPS32 + KS_MODE_LITTLE_ENDIAN)

    else:
        raise QlErrorArch("Unknown arch defined in utils.py (debug output mode)")

    return ks